from re import compile
from typing import List, Any, Dict, Tuple, Union, Optional, Callable
from pydantic import BaseModel as BasePydanticModel
from bson import ObjectId
from pymongo import UpdateOne
from pymongo.errors import (
    ServerSelectionTimeoutError,
    AutoReconnect,
    NetworkTimeout,
    ConnectionFailure,
    WriteConcernError,
)

from .exceptions import MongoConnectionError

__all__ = (
    'ExtraQueryMapper',
    'chunk_by_length',
    'bulk_query_generator',
    'generate_operator_for_multiply_aggregations',
    'cached_classproperty',
)


class cached_classproperty(classmethod):
    def __init__(self, fget):
        self.obj = {}
        self.fget = fget

    def __get__(self, owner, cls):
        if cls in self.obj:
            return self.obj[cls]
        self.obj[cls] = self.fget(cls)
        return self.obj[cls]


class ExtraQueryMapper(object):
    def __init__(self, field_name: str):
        self.field_name = field_name

    def extra_query(self, extra_methods: List, values) -> Dict:
        if self.field_name == '_id':
            values = (
                [ObjectId(v) for v in values]
                if isinstance(values, list)
                else ObjectId(values)
            )
        if extra_methods:
            query: Dict = {self.field_name: {}}
            for extra_method in extra_methods:
                if extra_method == 'in':
                    extra_method = 'in_'
                elif extra_method == 'inc':
                    return self.inc(values)
                elif extra_method == 'unset':
                    return self.unset(values)
                query[self.field_name].update(getattr(self, extra_method)(values))
            return query
        return {}

    def in_(self, list_values: List) -> dict:
        if not isinstance(list_values, list):
            raise TypeError("values must be a list type")
        return {"$in": list_values}

    def regex(self, regex_value: str) -> dict:
        return {"$regex": regex_value}

    def regex_ne(self, regex_value: str) -> dict:
        return {"$not": compile(regex_value)}

    def ne(self, value: Any) -> dict:
        return {"$ne": value}

    def startswith(self, value: str) -> dict:
        return {"$regex": f"^{value}"}

    def not_startswith(self, value: str) -> dict:
        return {"$not": compile(f"^{value}")}

    def endswith(self, value: str) -> dict:
        return {"$regex": f"{value}$"}

    def not_endswith(self, value: str) -> dict:
        return {"$not": compile(f"{value}$")}

    def nin(self, list_values: List) -> dict:
        if not isinstance(list_values, list):
            raise TypeError("values must be a list type")
        return {"$nin": list_values}

    def exists(self, boolean_value: bool) -> dict:
        if not isinstance(boolean_value, bool):
            raise TypeError("boolean_value must be a bool type")
        return {"$exists": boolean_value}

    def type(self, bson_type) -> dict:
        return {"$type": bson_type}

    def search(self, search_text: str) -> dict:
        return {'$search': search_text}

    def all(self, query: Any) -> dict:
        return {'$all': query}

    def unset(self, value: Any) -> dict:
        return {"$unset": {self.field_name: value}}

    def gte(self, value: Any) -> dict:
        return {"$gte": value}

    def lte(self, value: Any) -> dict:
        return {"$lte": value}

    def gt(self, value: Any) -> dict:
        return {"$gt": value}

    def lt(self, value: Any) -> dict:
        return {"$lt": value}

    def inc(self, value: int) -> dict:
        if isinstance(value, int):
            return {'$inc': {self.field_name: value}}
        raise ValueError('value must be integer')

    def range(self, range_values: Union[List, Tuple]) -> dict:
        if len(range_values) != 2:
            raise ValueError("range must have 2 params")
        from_ = range_values[0]
        to_ = range_values[1]
        return {"$gte": from_, "$lte": to_}

    @cached_classproperty
    def methods(cls) -> list:
        methods = []
        for f in cls.__dict__:
            if f == 'in_':
                methods.append('in')
            elif not f.startswith('__') and f != 'extra_query':
                methods.append(f)
        return methods


def chunk_by_length(items: List, step: int):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(items), step):
        yield items[i : i + step]


def bulk_query_generator(
    requests: List,
    updated_fields: Optional[List] = None,
    query_fields: Optional[List] = None,
    upsert=False,
) -> List:
    data = []
    if updated_fields:
        for obj in requests:
            query = obj.data
            query['_id'] = ObjectId(query['_id'])
            update = {}
            for field in updated_fields:
                value = query.pop(field)
                update.update({field: value})
            data.append(UpdateOne(query, {'$set': update}, upsert=upsert))
    elif query_fields:
        for obj in requests:
            query = {}
            update = {}
            for field, value in obj.data.items():
                if field not in query_fields:
                    update.update({field: value})
                else:
                    query.update({field: value})
            data.append(UpdateOne(query, {'$set': update}, upsert=upsert))
    return data


def handle_and_convert_connection_errors(func: Callable) -> Any:
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (
            ServerSelectionTimeoutError,
            AutoReconnect,
            NetworkTimeout,
            ConnectionFailure,
            WriteConcernError,
        ) as e:
            raise MongoConnectionError(str(e))

    return wrapper


def generate_lookup_project_params(
    main_model: BasePydanticModel, reference_models: Dict[str, BasePydanticModel]
) -> Dict:
    project_param = {f: 1 for f in main_model.__fields__}
    project_param['_id'] = 1
    for as_, reference_model in reference_models.items():
        project_param.update(
            {f'{as_}.{f}': 1 for f in ['_id'] + list(reference_model.__fields__.keys())}
        )
    return project_param


def generate_name_field(name: Union[dict, str, None] = None):
    if isinstance(name, dict):
        return '|'.join(str(v) for v in name.values())
    return name


def sort_validation(
    sort: Optional[int] = None, sort_fields: Union[list, tuple, None] = None
) -> Tuple[Any, ...]:
    if sort is not None:
        if sort not in (1, -1):
            raise ValueError(f'invalid sort value must be 1 or -1 not {sort}')
        if not sort_fields:
            sort_fields = ('_id',)
    return sort, sort_fields



from json import dumps
from typing import Generator, List, Optional, Union, Any
from pydantic.main import ModelMetaclass


from .helpers import handle_and_convert_connection_errors


class QuerySet(object):
    def __init__(
        self,
        model: ModelMetaclass,
        data: Generator,
        reference_models: Optional[List[ModelMetaclass]] = None,
    ):
        self._data = data
        self._model = model
        self._reference_models = reference_models

    @handle_and_convert_connection_errors
    def __iter__(self):
        for obj in self._data:
            yield self._model.parse_obj(obj, self._reference_models)

    @property
    def data(self) -> List:
        return [obj.data for obj in self.__iter__()]

    @property
    def json(self) -> str:
        return dumps(self.data)

    @property
    def generator(self) -> Generator:
        return self.__iter__()

    @property
    def data_generator(self) -> Generator:
        return (obj.data for obj in self.__iter__())

    @property
    def list(self) -> List:
        return list(self.__iter__())

    def first(self) -> Any:
        return next(self.__iter__())

    def serialize(self, fields: Union[tuple, List]) -> List:
        return [obj.serialize(fields) for obj in self.__iter__()]

    def serialize_generator(self, fields: Union[tuple, List]) -> Generator:
        for obj in self.__iter__():
            yield obj.serialize(fields)

    def serialize_json(self, fields: Union[tuple, List]) -> str:
        return dumps(self.serialize(fields))



import os
from typing import Optional


def init_db_connection_params(
    connection_str: str,
    dbname: str,
    ssl: bool = False,
    max_pool_size: int = 100,
    ssl_cert_path: Optional[str] = None,
    server_selection_timeout_ms: int = 50000,
    connect_timeout_ms: int = 50000,
    socket_timeout_ms: int = 50000,
) -> None:
    os.environ['MONGODANTIC_CONNECTION_STR'] = connection_str
    os.environ['MONGODANTIC_DBNAME'] = dbname
    os.environ['MONGODANTIC_SSL'] = '1' if ssl else '0'
    os.environ['MONGODANTICPOOL_SIZE'] = str(max_pool_size)
    if server_selection_timeout_ms:
        os.environ['MONGODANTIC_SERVER_SELECTION_TIMEOUT_MS'] = str(
            server_selection_timeout_ms
        )
    if connect_timeout_ms:
        os.environ['MONGODANTIC_CONNECT_TIMEOUT_MS'] = str(connect_timeout_ms)
    if socket_timeout_ms:
        os.environ['MONGODANTIC_SOCKET_TIMEOUT_MS'] = str(socket_timeout_ms)
    if ssl_cert_path:
        os.environ['MONGODANTIC_SSL_CERT_PATH'] = ssl_cert_path



import copy
import warnings
from typing import TYPE_CHECKING

from .helpers import ExtraQueryMapper
from .exceptions import DuplicateQueryParamError

__all__ = ("Query", "LogicalCombination")


if TYPE_CHECKING:
    from .models import MongoModel


def parse_query(model: 'MongoModel', query: dict) -> dict:
    return model._validate_query_data(query)


class QueryNodeVisitor(object):
    """Base visitor class for visiting Query-object nodes in a query tree.
    """

    def prepare_combination(self, combination):
        """Called by LogicalCombination objects.
        """
        return combination

    def visit_query(self, query):
        """Called by (New)Query objects.
        """
        return query


class SimplificationVisitor(QueryNodeVisitor):
    def __init__(self, model=None):
        self.model = model

    def prepare_combination(self, combination):
        if combination.operation == combination.AND:
            # The simplification only applies to 'simple' queries
            if all(isinstance(node, Query) for node in combination.children):
                queries = [n.query for n in combination.children]
                query = self._query_conjunction(queries)
                return {"$and": query}

        return combination

    def _query_conjunction(self, queries):
        """Merges query dicts - effectively &ing them together.
        """
        combined_query = []
        for query in queries:
            query = parse_query(self.model, query)
            combined_query.append(copy.deepcopy(query))
        return combined_query


class QueryCompilerVisitor(QueryNodeVisitor):
    """Compiles the nodes in a query tree to a PyMongo-compatible query
    dictionary.
    """

    def __init__(self, model):
        self.model = model

    def prepare_combination(self, combination):
        operator = "$and"
        if combination.operation == combination.OR:
            operator = "$or"
        return {operator: combination.children}

    def visit_query(self, query):
        query = parse_query(self.model, query.query)
        return query


class QueryNode(object):
    """Base class for nodes in query trees."""

    AND = 0
    OR = 1

    def to_query(self, model):
        query = self.accept(SimplificationVisitor(model))
        if not isinstance(query, dict):
            query = query.accept(QueryCompilerVisitor(model))
        return query

    def accept(self, visitor):
        raise NotImplementedError

    def _combine(self, other, operation):
        """Combine this node with another node into a LogicalCombination
        object.
        """
        # If the other Query() is empty, ignore it and just use `self`.
        if getattr(other, "empty", True):
            return self

        # Or if this Q is empty, ignore it and just use `other`.
        if self.empty:
            return other

        return LogicalCombination(operation, [self, other])

    @property
    def empty(self):
        return False

    def __or__(self, other):
        return self._combine(other, self.OR)

    def __and__(self, other):
        return self._combine(other, self.AND)


class LogicalCombination(QueryNode):
    def __init__(self, operation, children):
        self.operation = operation
        self.children = []
        for node in children:
            # If the child is a combination of the same type, we can merge its
            # children directly into this combinations children
            if isinstance(node, LogicalCombination) and node.operation == operation:
                self.children += node.children
            else:
                self.children.append(node)

    def __repr__(self):
        op = " & " if self.operation is self.AND else " | "
        return "(%s)" % op.join([repr(node) for node in self.children])

    def __bool__(self):
        return bool(self.children)

    def accept(self, visitor):
        for i in range(len(self.children)):
            if isinstance(self.children[i], QueryNode):
                self.children[i] = self.children[i].accept(visitor)

        return visitor.prepare_combination(self)

    @property
    def empty(self):
        return not bool(self.children)

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__
            and self.operation == other.operation
            and self.children == other.children
        )


class Query(QueryNode):
    """A simple query object, used in a query tree to build up more complex
    query structures.
    """

    def __init__(self, **query):
        self.query = query

    def __repr__(self):
        return "Query(**%s)" % repr(self.query)

    def __bool__(self):
        return bool(self.query)

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.query == other.query

    def accept(self, visit):
        return visit.visit_query(self)

    @property
    def empty(self):
        return not bool(self.query)



import os
from typing import Optional
from pymongo import MongoClient, database

all = ('_DBConnection',)


class _DBConnection(object):
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, 'instance'):
            cls.instance = super(_DBConnection, cls).__new__(cls)
        return cls.instance

    def __init__(self):
        self.connection_string = os.environ.get('MONGODANTIC_CONNECTION_STR')
        self.db_name = os.environ.get('MONGODANTIC_DBNAME')
        self.max_pool_size = int(os.environ.get('MONGODANTIC_POOL_SIZE', 100))
        self.ssl = bool(int(os.environ.get('MONGODANTIC_SSL', 0)))
        self.ssl_cert_path = os.environ.get('MONGODANTIC_SSL_CERT_PATH')
        self.server_selection_timeout_ms = int(
            os.environ.get('MONGODANTIC_SERVER_SELECTION_TIMEOUT_MS', 50000)
        )
        self.connect_timeout_ms = int(
            os.environ.get('MONGODANTIC_CONNECT_TIMEOUT_MS', 50000)
        )
        self.socket_timeout_ms = int(
            os.environ.get('MONGODANTIC_SOCKET_TIMEOUT_MS', 60000)
        )
        self._mongo_connection = self.__init_mongo_connection()
        self._database = None

    def __init_mongo_connection(self) -> MongoClient:
        connection_params = dict(
            connect=False,
            serverSelectionTimeoutMS=self.server_selection_timeout_ms,
            maxPoolSize=self.max_pool_size,
            connectTimeoutMS=self.connect_timeout_ms,
            socketTimeoutMS=self.socket_timeout_ms,
            retryWrites=False,
            retryReads=False,
        )
        if self.ssl:
            connection_params['tlsCAFile'] = self.ssl_cert_path
            connection_params['tlsAllowInvalidCertificates'] = self.ssl
        return MongoClient(self.connection_string, **connection_params)

    def _reconnect(self):
        self._mongo_connection = self.__init_mongo_connection()
        return self

    def get_database(self) -> database.Database:
        if hasattr(self, '_database') and self._database:
            return self._database
        self._database = self._mongo_connection.get_database(self.db_name)
        return self._database




from typing import Optional, TYPE_CHECKING

from .exceptions import ValidationError
from .helpers import generate_lookup_project_params

if TYPE_CHECKING:
    from .models import MongoModel


__all__ = ('Lookup', 'LookupCombination', 'Sum', 'Avg', 'Min', 'Count', 'Max')


class BasicDefaultAggregation(object):
    def __init__(self, field: str):
        self.field = field

    @property
    def _operation(self) -> str:
        raise NotImplementedError('implement _operation')

    def _validate_field(self, mongo_model: 'MongoModel'):
        if self.field not in mongo_model.__fields__ and self.field != '_id':
            raise ValidationError(
                f'{self.field} not in {mongo_model.__name__} field, field must be one of {list(mongo_model.__fields__.keys())}'
            )

    def _aggregate_query(self, mongo_model: 'MongoModel') -> dict:
        self._validate_field(mongo_model)
        query = {
            f'{self.field}__{self._operation}': {
                f'${self._operation}': f'${self.field}'
            }
        }
        return query


class Sum(BasicDefaultAggregation):
    @property
    def _operation(self) -> str:
        return 'sum'


class Max(BasicDefaultAggregation):
    @property
    def _operation(self) -> str:
        return 'max'


class Min(BasicDefaultAggregation):
    @property
    def _operation(self) -> str:
        return 'min'


class Avg(BasicDefaultAggregation):
    @property
    def _operation(self) -> str:
        return 'avg'


class Count(BasicDefaultAggregation):
    @property
    def _operation(self) -> str:
        return 'count'

    def _aggregate_query(self, mongo_model: 'MongoModel') -> dict:
        self._validate_field(mongo_model)
        query = {
            "_id": f'${self.field}' if self.field != '_id' else None,
            f'count': {f'$sum': 1},
        }
        return query


class LookupCombination(object):
    def __init__(self, lookup: list):
        self.children: list = []
        for node in lookup:
            if node in self.children:
                continue
            elif isinstance(node, LookupCombination):
                self.children.extend(node.children)
            else:
                self.children.append(node)

    def __repr__(self):
        return " AND ".join([repr(node) for node in self.children])

    def accept(self, main_model: 'MongoModel', project: Optional[dict] = None) -> tuple:
        accepted_lookup = []
        reference_models = {}
        for node in self.children:
            accepted_lookup.extend(node.to_query(main_model))
            reference_models[node.as_] = node.from_collection
        if not project:
            project = generate_lookup_project_params(main_model, reference_models)
        accepted_lookup.append({'$project': project})
        return accepted_lookup, reference_models


class Lookup(object):
    def __init__(
        self,
        from_collection: 'MongoModel',
        local_field: str,
        foreign_field: str,
        as_: Optional[str] = None,
        with_unwind: bool = False,
        preserve_null_and_empty_arrays: bool = False,
    ):
        self.from_collection = from_collection
        self.local_field = local_field
        self.foreign_field = foreign_field
        self.as_ = as_ if as_ else self.from_collection.collection_name
        self.with_unwind = with_unwind
        self.preserve_null_and_empty_arrays = preserve_null_and_empty_arrays

    def to_query(self, main_model: 'MongoModel') -> list:
        query = [
            {
                '$lookup': {
                    'localField': self._validate_local_field(main_model),
                    'from': self.from_collection.collection_name,
                    'foreignField': self.foreign_field,
                    'as': self.as_,
                }
            }
        ]
        if self.with_unwind:
            query.append(
                {
                    '$unwind': {
                        'path': f'${self.as_}',
                        'preserveNullAndEmptyArrays': self.preserve_null_and_empty_arrays,
                    }
                }
            )
        return query

    def _validate_local_field(self, main_model: 'MongoModel') -> str:
        if (
            self.local_field
            not in set(
                self.from_collection.__fields__.keys() | main_model.__fields__.keys()
            )
            and self.local_field != '_id'
        ):
            raise AttributeError('invalid local_field')
        return self.local_field

    def _combine(self, other) -> LookupCombination:
        return LookupCombination([self, other])

    def accept(self, main_model: 'MongoModel', project: Optional[dict] = None) -> tuple:
        return LookupCombination([self]).accept(main_model, project)

    def __and__(self, other):
        return self._combine(other)

    def __repr__(self):
        return f'Lookup(from_collection={self.from_collection.__name__.lower()}, local_field={self.local_field}, foreign_field={self.foreign_field}, as={self.as_}'




import unittest
from bson import ObjectId

from mongodantic.models import MongoModel
from mongodantic import init_db_connection_params
from mongodantic.session import Session


class TestQueriesWithInners(unittest.TestCase):
    def setUp(self):
        init_db_connection_params("mongodb://127.0.0.1:27017", "test")

        class InnerTicket(MongoModel):
            name: str
            position: int
            config: dict
            sign: int = 1
            type_: str = 'ga'

            class Config:
                excluded_query_fields = ('sign', 'type')

        InnerTicket.querybuilder.drop_collection(force=True)
        self.InnerTicket = InnerTicket

    def create_documents(self):
        self.InnerTicket.querybuilder.insert_one(
            name='first', position=1, config={'url': 'localhost', 'username': 'admin'},
        )
        self.InnerTicket.querybuilder.insert_one(
            name='second',
            position=2,
            config={'url': 'google.com', 'username': 'staff'},
        )
        self.InnerTicket.querybuilder.insert_one(
            name='third',
            position=3,
            config={'url': 'yahoo.com', 'username': 'trololo'},
        )

    def test_update_many(self):
        self.create_documents()
        updated = self.InnerTicket.querybuilder.update_many(
            position__range=[2, 3], name__ne='hhh', config__url__set='test.io'
        )
        assert updated == 2
        last = self.InnerTicket.querybuilder.find_one(sort=-1)
        assert last.config['url'] == 'test.io'

    def test_inner_find_one(self):
        self.create_documents()
        data = self.InnerTicket.querybuilder.find_one(config__url__startswith='goo')
        assert data.name == 'second'

    def test_inner_update_one(self):
        self.create_documents()
        updated = self.InnerTicket.querybuilder.update_one(
            config__url__startswith='goo', config__url__set='test.io'
        )
        assert updated == 1
        data = self.InnerTicket.querybuilder.find_one(config__url__startswith='test')
        assert data.name == 'second'



from typing import Optional
from pymongo.collection import Collection

from .db import _DBConnection
from .helpers import cached_classproperty
from .querybuilder import QueryBuilder
from .types import ObjectIdStr


class ModelMixin(object):
    _id: Optional[ObjectIdStr] = None

    @cached_classproperty
    def querybuilder(cls):
        querybuilder = QueryBuilder()
        querybuilder.add_model(cls)
        return querybuilder

    @property
    def pk(self):
        return self._id

    def __name__(self):
        return super().__name__()


class DBConnectionMixin(object):
    _connection = _DBConnection()

    @classmethod
    def _reconnect(cls):
        cls._connection = cls._connection._reconnect()

    @classmethod
    def get_database(cls):
        return cls._connection.get_database()

    @classmethod
    def set_collection_name(cls) -> str:
        return cls.__name__.lower()

    @classmethod
    def get_collection(cls) -> Collection:
        db = cls.get_database()
        return db.get_collection(cls.collection_name)

    @cached_classproperty
    def collection_name(cls):
        return cls.set_collection_name()

    @cached_classproperty
    def collection(cls):
        return cls.get_collection()



from typing import Any
from contextlib import ContextDecorator

from .db import _DBConnection

__all__ = ("Session",)


class Session(ContextDecorator):
    _connection = _DBConnection()

    def __init__(self):
        self._session = self._connection._mongo_connection.start_session()

    def __enter__(self):
        return self._session

    def __exit__(self, *exc):
        return self.close()

    def close(self):
        return self._session.end_session()



import pymongo
import unittest
import pytest
from mongodantic.models import MongoModel
from mongodantic import init_db_connection_params
from mongodantic.exceptions import MongoIndexError


class TestIndexOperation(unittest.TestCase):
    def setUp(self):
        init_db_connection_params("mongodb://127.0.0.1:27017", "test")

        class Ticket(MongoModel):
            name: str
            position: int
            config: dict

        Ticket.querybuilder.drop_collection(force=True)
        self.Ticket = Ticket

    def test_add_index(self):
        result = self.Ticket.querybuilder.add_index('position', 1)
        assert result == 'index with name - position_1 created.'

        with pytest.raises(MongoIndexError):
            result = self.Ticket.querybuilder.add_index('position', 1)

    def test_check_indexes(self):
        self.test_add_index()
        result = self.Ticket.querybuilder.check_indexes()
        assert result == [
            {'name': '_id_', 'key': {'_id': 1}},
            {'name': 'position_1', 'key': {'position': 1}},
        ]

    def test_drop_index(self):
        self.test_add_index()
        with pytest.raises(MongoIndexError):
            result = self.Ticket.querybuilder.drop_index('position1111', 1)

        result = self.Ticket.querybuilder.drop_index('position', 1)
        assert result == 'position_1 dropped.'



from os import path
from io import open
from setuptools import setup, find_packages

from mongodantic import __version__


def read(f):
    return open(f, "r").read()


setup(
    name="mongodantic",
    version=__version__,
    packages=find_packages(exclude=("tests",)),
    install_requires=["pydantic>=1.3,<2", "pymongo>=3.10.1,<3.11",],
    description="Mongo ODM, based on pydantic",
    author="bzdvdn",
    author_email="bzdv.dn@gmail.com",
    url="https://github.com/bzdvdn/mongodantic",
    license="MIT",
    python_requires=">=3.6",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
)



from bson import ObjectId
from bson.errors import InvalidId


class ObjectIdStr(str):
    """Field for validate string like ObjectId"""

    type_ = ObjectId
    required = False
    default = None
    validate_always = False
    alias = ''

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(
        cls, v,
    ):
        if isinstance(v, ObjectId):
            return v
        else:
            try:
                return ObjectId(str(v))
            except InvalidId:
                raise ValueError(f"invalid ObjectId - {v}")



class BaseMongodanticException(Exception):
    pass


class NotDeclaredField(BaseMongodanticException):
    def __init__(self, field_name: str, fields: list, *args):
        self.field_name = field_name
        self.fields = fields
        super().__init__(*args)

    def __str__(self):
        return f"This field - {self.field_name} not declared in {self.fields}"


class InvalidArgument(BaseMongodanticException):
    pass


class ValidationError(BaseMongodanticException):
    pass


class MongoIndexError(BaseMongodanticException):
    pass


class MongoConnectionError(BaseMongodanticException):
    pass


class DuplicateQueryParamError(BaseMongodanticException):
    pass


class InvalidArgsParams(BaseMongodanticException):
    def __str__(self):
        return 'Arguments must be Query objects'



from json import dumps
from typing import Dict, Any, Union, Optional, List, Tuple
from pymongo.client_session import ClientSession
from bson import ObjectId
from pydantic.main import ModelMetaclass
from pydantic import BaseModel as BasePydanticModel
from pydantic import validator

from .mixins import DBConnectionMixin, ModelMixin
from .types import ObjectIdStr
from .exceptions import (
    NotDeclaredField,
    ValidationError,
    InvalidArgsParams,
)
from .helpers import ExtraQueryMapper
from .queryset import QuerySet
from .logical import LogicalCombination, Query


__all__ = ('MongoModel', 'QuerySet', 'Query')


class BaseModel(DBConnectionMixin, ModelMixin, BasePydanticModel):
    class Config:
        excluded_query_fields = ()

    def __setattr__(self, key, value):
        if key in self.__fields__:
            return super().__setattr__(key, value)
        self.__dict__[key] = value
        return value

    @classmethod
    def parse_obj(
        cls, data: Any, reference_models: Dict[Any, 'BaseModel'] = {},
    ) -> Any:
        obj = super().parse_obj(data)
        if '_id' in data:
            obj._id = data['_id']
        if reference_models:
            obj = cls.__set_reference_fields(obj, data, reference_models)
        return obj

    @classmethod
    def __set_reference_fields(
        cls, obj: 'BaseModel', data: Dict, reference_models: Dict[Any, 'BaseModel'],
    ) -> 'BaseModel':
        for name_as, ref in reference_models.items():
            data = data[name_as]
            if isinstance(data, dict):
                ref_obj = ref.parse_obj(data)
            else:
                ref_obj = [ref.parse_obj(d) for d in data]
            setattr(obj, f'{name_as}', ref_obj)
        return obj

    @classmethod
    def __validate_field(cls, field: str) -> bool:
        if field not in cls.__fields__ and field != '_id':
            raise NotDeclaredField(field, list(cls.__fields__.keys()))
        elif field in cls.Config.excluded_query_fields:
            return False
        return True

    @classmethod
    def _parse_extra_params(cls, extra_params: tuple) -> tuple:
        field_param, extra = [], []
        methods = ExtraQueryMapper.methods
        for param in extra_params:
            if param in methods:
                extra.append(param)
            else:
                field_param.append(param)
        return field_param, extra

    @classmethod
    def _validate_query_data(cls, query: Dict) -> Dict:
        data = {}
        for field, value in query.items():
            field, *extra_params = field.split("__")
            inners, extra_params = cls._parse_extra_params(extra_params)
            if not cls.__validate_field(field):
                continue
            _dict = ExtraQueryMapper(field).extra_query(extra_params, value)
            if _dict:
                value = _dict[field]
            elif field == '_id':
                value = ObjectId(value)
            else:
                value = cls.__validate_value(field, value) if not inners else value
            if inners:
                field = f'{field}.{".".join(i for i in inners)}'
            data[field] = value
        return data

    @classmethod
    def __validate_value(cls, field_name: str, value: Any) -> Any:
        field = cls.__fields__[field_name]
        if isinstance(field, ObjectIdStr):
            try:
                value = field.validate(value)
            except ValueError as e:
                error_ = e
        else:
            value, error_ = field.validate(value, {}, loc=field.alias, cls=cls)
        if error_:
            raise ValidationError([error_], type(value))
        return value

    @classmethod
    def _check_query_args(
        cls, logical_query: Union[Query, LogicalCombination, None] = None
    ) -> Dict:
        if not isinstance(logical_query, (LogicalCombination, Query)):
            raise InvalidArgsParams()
        return logical_query.to_query(cls)

    @classmethod
    def _start_session(cls) -> ClientSession:
        return cls._connection._mongo_connection.start_session()

    @classmethod
    def sort_fields(cls, fields: Union[tuple, list, None]) -> None:
        if fields:
            new_sort = {field: cls.__fields__[field] for field in fields}
            cls.__fields__ = new_sort

    @property
    def data(self) -> Dict:
        data = self.dict()
        if '_id' in data:
            data['_id'] = data['_id'].__str__()
        return data


class MongoModel(BaseModel):
    def save(
        self,
        updated_fields: Union[Tuple, List] = [],
        session: Optional[ClientSession] = None,
    ) -> Any:
        if self._id is not None:
            data = {'_id': ObjectId(self._id)}
            if updated_fields:
                if not all(field in self.__fields__ for field in updated_fields):
                    raise ValidationError('invalid field in updated_fields')
            else:
                updated_fields = tuple(self.__fields__.keys())
            for field in updated_fields:
                data[f'{field}__set'] = getattr(self, field)
            self.querybuilder.update_one(**data, session=session)
            return self
        data = {
            field: value
            for field, value in self.__dict__.items()
            if field in self.__fields__
        }
        object_id = self.querybuilder.insert_one(**data, session=session)
        self._id = object_id.__str__()
        return self

    def delete(self, session: Optional[ClientSession] = None) -> None:
        self.querybuilder.delete_one(_id=ObjectId(self._id), session=session)

    def drop(self, session: Optional[ClientSession] = None) -> None:
        return self.delete(session)

    def serialize(self, fields: Union[tuple, list]) -> dict:
        data = self.dict(include=set(fields))
        return {f: data[f] for f in fields}

    def serialize_json(self, fields: Union[tuple, list]) -> str:
        return dumps(self.serialize(fields))

    def __hash__(self):
        if self.pk is None:
            raise TypeError("MongoModel instances without _id value are unhashable")
        return hash(self.pk)



import unittest
import pytest
from bson import ObjectId
from uuid import uuid4, UUID

from mongodantic.models import MongoModel
from mongodantic import init_db_connection_params
from mongodantic.session import Session
from mongodantic.exceptions import ValidationError


class TestBasicOperation(unittest.TestCase):
    def setUp(self):
        init_db_connection_params("mongodb://127.0.0.1:27017", "test")

        class User(MongoModel):
            id: UUID
            name: str
            email: str

            class Config:
                excluded_query_fields = ('sign', 'type')

        User.querybuilder.drop_collection(force=True)
        self.User = User

    def test_raw_insert_one(self):
        with pytest.raises(ValidationError):
            result = self.User.querybuilder.raw_query(
                'insert_one', {'id': uuid4(), 'name': {}, 'email': []}
            )
        result = self.User.querybuilder.raw_query(
            'insert_one', {'id': uuid4(), 'name': 'first', 'email': 'first@mail.ru'}
        )
        assert isinstance(result.inserted_id, ObjectId)

    def test_raw_insert_many(self):
        with pytest.raises(ValidationError):
            result = self.User.querybuilder.raw_query(
                'insert_many', [{'id': uuid4(), 'name': {}, 'email': []}]
            )
        result = self.User.querybuilder.raw_query(
            'insert_many', [{'id': uuid4(), 'name': 'first', 'email': 'first@mail.ru'}]
        )
        assert len(result.inserted_ids) == 1

    def test_raw_find_one(self):
        self.test_raw_insert_one()
        result = self.User.querybuilder.raw_query('find_one', {'name': 'first'})
        assert result['name'] == 'first'
        assert result['email'] == 'first@mail.ru'

    def test_raw_update_one(self):
        self.test_raw_insert_one()
        with pytest.raises(ValidationError):
            result = self.User.querybuilder.raw_query(
                'update_one', [{'id': uuid4(), 'name': {}, 'email': []}]
            )
        result = self.User.querybuilder.raw_query(
            'update_one', raw_query=({'name': 'first'}, {'$set': {'name': 'updated'}})
        )

        assert result.modified_count == 1

        modifed_result = self.User.querybuilder.find_one(email='first@mail.ru')
        assert modifed_result.name == 'updated'
